###############################################################
# pytest -v --capture=no tests/test_config.py
# pytest -v  tests/test_config.py
# pytest -v --capture=no  tests/test_config..py::Test_config::<METHODNAME>
###############################################################

import os
from pprint import pprint

import pytest
from yamldb.YamlDB import YamlDB

from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.util import HEADING
from cloudmesh.common.util import readfile
from cloudmesh.common.systeminfo import os_is_windows
from cloudmesh.common.util import path_expand

if os_is_windows():
    filename = "~/.cloudmesh/cloudmesh.yaml"
else:
    filename = "/tmp/cloudmesh.yaml"

filename = path_expand(filename)

data = {
    "a": 1,
    "b": "text",
    "c": {
        "c1": "c",
        "c2": "cc",
    },
}


@pytest.mark.incremental
class TestConfig:
    def test_01_init(self):
        HEADING()
        StopWatch.start("init")
        db = YamlDB(data=data, filename=filename)
        StopWatch.stop("init")
        db.save()
        content = readfile(filename)
        print(db.data)
        print(content)

        assert "a" in db.data
        assert db.data["c"]["c1"] == "c"
        assert os.path.isfile(filename)

    def test_02_read(self):
        HEADING()
        StopWatch.start("read")
        db = YamlDB(filename=filename)
        db.load(filename=filename)
        StopWatch.stop("read")
        assert type(db["a"]) == int
        assert "a" in db.data

    def test_03_dict(self):
        HEADING()
        db = YamlDB(filename=filename)
        StopWatch.start("dict")
        result = db.dict()
        StopWatch.stop("dict")
        pprint(result)
        print(db)
        print(type(db.data))
        assert "a" in result
        assert "c.c1" in db

    def test_04_set_nested(self):
        HEADING()
        StopWatch.start("set")
        db = YamlDB(filename=filename)
        db["cloudmesh.test.nested"] = "Gregor"
        StopWatch.stop("set")
        print(db["cloudmesh.test.nested"])
        assert db["cloudmesh.test.nested"] == "Gregor"

    def test_05_doesnotexist_key(self):
        HEADING()
        db = YamlDB(filename=filename)

        key = "cloudmesh.doesnotexist"
        StopWatch.start(f"db[{key}]")
        try:
            db[key]
            result = True
        except:
            result = False
        StopWatch.stop(f"db[{key}]")

        assert result is False

    def test_06_get_create(self):
        HEADING()
        db = YamlDB(filename=filename)

        key = "cloudmesh.doesnotexist"
        StopWatch.start(f"db.get({key})")
        value = db.get(key, default="Hallo")
        StopWatch.stop(f"db.get({key})")

        print(db)

        assert value == "Hallo"

    def test_07_search(self):
        db = YamlDB(data=data, filename=filename)

        StopWatch.start("search")
        r = db.search("c.*")
        StopWatch.stop("search")
        pprint(r)
        assert ["c", "cc"] == r
        assert db is not None

    def test__08_set_does_not_exist(self):
        HEADING()
        db = YamlDB(filename=filename)
        key = "cloudmesh.doesnotexist"

        db["cloudmesh.test.a"] = "aa"
        print(db)

        StopWatch.start("delete")
        db.delete("a")  # does not throw exception
        print(db)

        assert "cloudmesh.test" in db
        db.delete("cloudmesh.test")
        print(db)
        assert "cloudmesh.test" not in db  # this is a bug

        del db[key]
        StopWatch.stop("delete")
        assert "a" not in db
        assert "cloudmesh.doesnotexist" not in db

    def test_09_StopWatch(self):
        StopWatch.benchmark(sysinfo=False)



"""A small set of utility functions for file operations."""


def readfile(filename, mode="r"):
    """Reads the content of a file.

    Args:
        filename (str): The filename.
        mode (str): The file mode. Default is 'r'.

    Returns:
        str: The content of the file.

    Raises:
        ValueError: If an incorrect mode is provided.
    """
    if mode not in ["r", "rb"]:
        raise ValueError(f"Incorrect mode: expected 'r' or 'rb', given {mode}")

    with open(filename, mode) as file:
        content = file.read()
    return content


def writefile(filename, content):
    """
    Writes the content into the file.

    Args:
        filename (str): The filename.
        content (str): The content to be written.

    Returns:
        None
    """
    with open(filename, "w") as outfile:
        outfile.write(content)
        outfile.truncate()



"""from yamldb import YamlDB
from yamldb import Query

db = YamlDB('path/to/db.yaml')
user = Query()
db.insert({'name': 'Gregor', 'age': 111})
db.search("[?name=='Gregor']")

[{'name': 'Gregor', 'age': 111}]

"""
# pylint: disable=C0103,W0107,W0702
import os

import jmespath
import oyaml as yaml
from collections.abc import MutableMapping
from contextlib import suppress
from cloudmesh.common import dotdict


class YamlDB:
    """The YamlBD class uses a file based yaml file as its database backend."""

    def __init__(self, *, data=None, filename="yamldb.yml", backend=":file:"):
        """
        Initializes the YamlDB object.

        Args:
            data (dict, optional): The initial data to be stored in the database. Defaults to None.
            filename (str, optional): The name of the file to save the database. Defaults to "yamldb.yml".
            backend (str, optional): The backend storage type. Must be either ":file:" or ":memory:". Defaults to ":file:".

        Raises:
            ValueError: If the backend is not ":file:" or ":memory:".
            ValueError: If loading the database fails.

        """
        if backend not in [":file:", ":memory:"]:
            raise ValueError("backend must be :file: or :memory:")
        self.backend = backend
        self.filename = filename

        if os.path.exists(filename):
            if data is not None:
                self.data = data
                self.save(filename=filename)
            else:
                self.load(filename=filename)
        else:
            self._create_dir(filename)
            self.data = data if data is not None else {}
            self.save(filename=filename)

    def _create_dir(self, filename):
        """
        Create a directory for the given filename if it doesn't already exist.

        Args:
            filename (str): The path to the file.

        Raises:
            ValueError: If the directory cannot be created.

        """
        directory = os.path.dirname(filename)
        if directory is not None and directory != "":
            if not os.path.isdir(directory):
                try:
                    os.makedirs(directory, exist_ok=True)
                except OSError as e:
                    raise ValueError(
                        f"YAMLDB: could not create directory={directory}"
                    ) from e

    def print_dictionary(self, dic, indent=0):
        """
        Prints the contents of a dictionary recursively.

        Parameters:
            dic (dict): The dictionary to be printed.
            indent (int): The number of tabs to be used for indentation.

        Returns:
            None
        """
        if len(dic) == 0:
            print("\t" * indent + "{}")
        for key, value in dic.items():
            print("\t" * indent + str(key))
            if isinstance(value, dict):
                self.print_dictionary(value, indent + 1)
            else:
                print("\t" * (indent + 1) + str(type(value)))

    def get_keys(self, d=None, keys_list=None, parent=""):
        """
        Recursively retrieves all keys in the YAML data.

        Args:
            d (dict, optional): The dictionary to traverse. Defaults to None.
            keys_list (list, optional): The list to store the keys. Defaults to None.
            parent (str, optional): The parent key. Defaults to "".

        Returns:
            list: A list of all keys in the YAML data.
        """
        if keys_list is None:
            keys_list = []
            keys_list = list(self.data.keys())
        if d is None:
            d = dict(self.data)
        for key, value in d.items():
            if isinstance(d[key], dict):
                self.get_keys(d[key], keys_list, parent=key)
            else:
                keys_list.append(f"{parent}.{key}")
        return keys_list

    def keys(self):
        """
        Returns a list of all the keys in the YamlDB object.
        """
        return self.get_keys()

    def __iter__(self):
        """Iterates over the first level of the keys.

        Returns:
            An iterator object that iterates over the keys in the first level of the data.
        """
        return iter(self.data)

    def __contains__(self, key):
        """Checks if the key is included in the db

        Args:
            key: The key to check for existence in the database.

        Returns:
            bool: True if the key is found in the database, False otherwise.
        """
        found = True
        try:
            self.__getitem__(key)
        except:
            found = False
        return found

    def clear(self):
        """Removes all data"""
        self.data.clear()
        self.flush()

    def load(self, filename=None):
        """Loads the data from the specified filename

        Args:
            filename (str): The path to the file to load the data from. If not provided, the default filename will be used.

        Returns:
            None

        """
        prefix = os.path.dirname(filename)
        if prefix is not None and prefix != "":
            if not os.path.exists(prefix):
                os.makedirs(prefix)

        name = filename or self.filename

        if os.path.exists(name):
            with open(name, "rb") as dbfile:
                self.data = yaml.safe_load(dbfile) or dict()

    def update(self, filename=None):
        """Inserts the data from the specified filename

        Args:
            filename (str): The path to the file containing the data to be inserted.

        Returns:
            None
        """
        # prefix = os.path.dirname(filename)
        # if not os.path.exists(prefix):
        #    os.makedirs(prefix)
        # the entry must have an id which is used as the name for the entry

        if os.path.exists(filename):
            with open(filename, "rb") as dbfile:
                data = yaml.safe_load(dbfile) or dict()
                id = data["id"] or "unknown"
                if id in ["unknown", "MISSING"]:
                    print(f"Error: id not found for {filename}")
                d = {id: data}
                self.data.update(d)

    def save(self, filename=None, indent=None):
        """
        Saves the data to the specified filename.

        Args:
            filename (str): The name of the file to save the data to. If not provided, the current filename will be used.
            indent (int): The number of spaces to use for indentation. If not provided, the default indentation will be used.

        Returns:
            None
        """
        name = filename or self.filename
        with open(name, "w") as stream:
            yaml.dump(self.data, stream, default_flow_style=False, indent=indent)

    def flush(self):
        """
        Saves the data to the file specified when the DB was loaded.

        Returns:
            None

        """
        if self.backend == ":file:":
            self.save()

    def close(self):
        """
        Close the DB without flushing the current content.

        Returns:
            None
        """
        self.flush()

    def dict(self):
        """Returns the dictionary representation of the data.

        Returns:
            dict: The dictionary representation of the data.
        """
        return self.data

    def yaml(self):
        """Returns the YAML representation of the data.

        Returns:
            str: The YAML string representation of the data.
        """
        return str(self.data)

    def __len__(self):
        """Return the number of elements in the top level.

        Returns:
            int: The number of elements in the top level.
        """
        len(self.data)

    def __setitem__(self, key, value):
        """Sets the item given at the key to value

        Args:
            key: The key to set the value for.
            value: The value to be set.

        Returns:
            None
        """
        self.set(key, value)

    def set(self, key, value):
        """A helper function for setting the default cloud in the config without
        a chain of `set()` calls.

        Usage:
            value = db.set('a.b.c.d', 'value')

        Args:
            key: A string representing the value's path in the config.
            value: value to be set.

        Raises:
            ValueError: If the key is not found in the yaml file or if there is an unknown error.
        """
        try:
            if value.lower() in ["true", "false"]:
                value = value.lower() == "true"
        except:
            pass

        try:
            if "." in key:
                keys = key.split(".")
                #
                # create parents
                #
                parents = keys[:-1]
                location = self.data
                for parent in parents:
                    if parent not in location:
                        location[parent] = {}
                    location = location[parent]
                #
                # create entry
                #
                location[keys[-1]] = value
            else:
                self.data[key] = value

        except KeyError:
            raise ValueError(
                f"The key '{key}' could not be found in the yaml file '{self.filename}'"
            )
        except Exception as e:
            print(e)
            raise ValueError("unknown error")

        self.flush()

    def _delete_keys_from_dict(self, data, keys):
        """
        Recursively deletes keys from a nested dictionary.

        Args:
            data (dict): The dictionary from which keys will be deleted.
            keys (list): A list of keys to be deleted.

        Returns:
            None
        """
        for key in keys:
            with suppress(KeyError):
                print("DELETE", key, keys)
                del data[key]
        for value in data.values():
            if isinstance(value, MutableMapping):
                self._delete_keys_from_dict(value, keys)

    def delete(self, item):
        """Deletes an item from the dict. The key is . separated
        use it as follows get("a.b.c")

        Args:
            item (str): The key of the item to be deleted.

        Returns:
            None
        """
        try:
            if "." in item:
                keys = item.split(".")
                d = self.data
                for key in keys[:-1]:
                    d = d[key]
                    print("K", key, d)
                del_key = keys[-1]
                del d[del_key]
                self._delete_keys_from_dict(self.data, keys)
            else:
                del self.data[item]
                return
        except Exception as e:
            print(e)
            # raise ValueError("unknown error")

    def __delitem__(self, key):
        """
        Delete an item from the YamlDB instance.

        Args:
            key: The key of the item to delete.

        Returns:
            None
        """
        self.delete(key)

    def __getitem__(self, item):
        """gets an item from the dict. The key is . separated
        use it as follows get("a.b.c")

        Args:
            item (str): The key to retrieve from the dictionary.

        Returns:
            Any: The value associated with the given key.

        Raises:
            KeyError: If the key is not found in the yaml file.
            ValueError: If an unknown error occurs.
        """
        try:
            if "." in item:
                keys = item.split(".")
            else:
                return self.data[item]
            element = self.data[keys[0]]
            for key in keys[1:]:
                element = element[key]
        except KeyError:
            raise KeyError(
                f"The key '{item}' could not be found in the yaml file '{self.filename}'"
            )
        except Exception as e:
            print(e)
            raise ValueError("unknown error")
        return element

    def get(self, key, default=None):
        """Gets the value based on the key.

        Args:
            key (str): A string representing the value's path in the config.
            default (Any, optional): The default value to return if the key is not found. Defaults to None.

        Returns:
            Any: The value associated with the key, or the default value if the key is not found.
        """
        try:
            return self.data[key]
        except KeyError:
            self[key] = default
            return default

    def search(self, query):
        """
        Searches the data using JMESPath query language.

        Args:
            query (str): The JMESPath query string.

        Returns:
            dict: A dictionary containing the search result.
        """
        return jmespath.search(query, self.data)

    def __str__(self):
        """Returns the YAML representation of the data as a string.

        Returns:
            str: The YAML representation of the data.
        """
        if self.data is None:
            return ""

        return yaml.dump(self.data, default_flow_style=False, indent=2)



###############################################################
# pytest -v --capture=no tests/test_delete.py
# pytest -v  tests/test_delete.py
# pytest -v --capture=no  tests/test_delete..py::test_delete::<METHODNAME>
###############################################################

import os
from pprint import pprint

import pytest
from yamldb.YamlDB import YamlDB
import yaml

from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.util import HEADING
from cloudmesh.common.util import readfile
from cloudmesh.common.systeminfo import os_is_windows
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import banner

if os_is_windows():
    filename = "~/.cloudmesh/cloudmesh.yaml"
else:
    filename = "/tmp/cloudmesh.yaml"

filename = path_expand(filename)


data = {
    "floor": {
        "key": "value",
        "key_a": "value_a",
        "key_b": "value_b"
    }
}

def create_data_file(filename, data):
    with open(filename, 'w') as file:
        yaml.dump(dict(data), file)

@pytest.mark.incremental
class TestConfig:

    def test_delete(self):
        HEADING()
        global data
        print (data)
        assert os.path.isfile(filename)

        StopWatch.start("init")
        db = YamlDB(data=dict(data), filename=filename)
        StopWatch.stop("init")
        db.save()
        content = readfile(filename)
        banner("content")
        print(content)
        banner("db.data")
        print(db)
        banner("value")
        print (db["floor.key"])
        assert db["floor.key"] == "value"

        keys = db.keys()
        assert len(keys) == 4
        print (keys)

        # banner("delete")
        # StopWatch.start("delete")
        # db.delete("floor.key")
        # StopWatch.stop("delete")
        # assert "floor.key" not in db
        # banner("db.data after delete")
        # print(db)

class g:

    def test_read_from_file(self):
        HEADING()
        global data
        print("DDD", data)
        self.filename = "tests/test.yaml"

        create_data_file(self.filename, data)

        banner(self.filename)
        content = readfile(self.filename)
        print(content)


        print(self.filename)
        db = YamlDB(filename=self.filename)
        db.save()

        banner("db.data")
        print(db)
        banner("value")
        print(db["floor.key"])

        assert db["floor.key"] == "value"

        banner("delete")
        StopWatch.start("delete")
        db.delete("floor.key")
        db.save()
        StopWatch.stop("delete")
        assert "floor.key" not in db
        banner("db.data after delete")
        print(db)

        banner(self.filename)
        content = readfile(self.filename)
        print (content)
    #

    def test_delete_variants(self):
        HEADING()
        self.filename = "tests/test.yaml"

        create_data_file(self.filename, data)


        db = YamlDB(filename=self.filename)
        db["floor.a"] = "value_a"
        db["floor.b"] = "value_b"
        db.save()

        banner("db values")
        print("BBBB", db)

        for key in ["floor.a", "floor.b"]:
            print(f"get   : {key}: {db[key]}")
            r = db.search("floor.a")
            print(f"search: {key}: {r}")



        db.delete("floor.a")

        db.save()
        print("AAAA", db.data)


        banner("db.data delete a")
        print(db)


        #
        # THIS DOES NOT YET WORK
        #
        # db.delete["floor"]
        # banner("db.data after delete")
        # print(db)

        #
        # THIS IS NOT YET TESTED
        #
        #try:
        #    db.delete["floor.key.nothing"]
        #except Exception as e:
        #    print (e)
        #    assert "could not be found" in str(e)
        #except TypeError:
        #    print ("UUU")

            # print (e)

        # banner ("AAA")
        # """
        #
        #
        # del db["floor"]
        # del db["floor.key"]
        # del ["floor.key.nothing"]
        # """

class aaa:

    def test_StopWatch(self):
        StopWatch.benchmark(sysinfo=False)



###############################################################
# pytest -v --capture=no tests/test_issues.py
# pytest -v  tests/test_issues.py
# pytest -v --capture=no  tests/test_issues..py::test_issues::<METHODNAME>
###############################################################

import os
from pprint import pprint

import pytest
from yamldb.YamlDB import YamlDB

from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.util import HEADING
from cloudmesh.common.util import readfile
from cloudmesh.common.systeminfo import os_is_windows
from cloudmesh.common.util import path_expand


@pytest.mark.incremental
class TestConfig:
    def test_issue_6_from_data(self):
        HEADING()
        data = YamlDB(filename="tmp_a.yaml", data={"bot": {"lang": "a"}})
        print(data)

        bot_lang = data["bot.lang"]
        print("Result:", bot_lang)
        assert bot_lang == "a"

    def test_issue_6_from_value(self):
        HEADING()
        data = YamlDB(filename="tmp_b.yaml")
        # data["bot.lang"] = "a"
        # print(data)
        # bot_lang = data["bot.lang"]
        # print("Result:", bot_lang)

    # def test_issue_6_from_case(self):
    #     locals = YamlDB(filename="tmp_locals.yaml")
    #     data = YamlDB(filename="tmp_data.yaml")
    #     data["bot.lang"] = "a"
    #     bot_lang = data["bot.lang"]
    #     assert bot_lang == "a"
